(*  *)
(*   Autor:  Ignacio Prez Martn.  Fecha: 15 de Mayo de 2002.   *)
(*   Msica: Roco Lpez Abelaira.                               *)
(*  *)

IMPLEMENTATION MODULE Serp;
FROM Storage IMPORT ALLOCATE, DEALLOCATE;
IMPORT Graf, Graph;

CONST
   COLOR_SERP = 02;
   NEGRO      = 00;
   MASC_SERP  = Graph.FillMaskType(011H, 077H, 0AAH, 0CCH,
                                   011H, 077H, 0AAH, 0CCH);
   NORMAL     = Graph.FillMaskType(0FFH, 0FFH, 0FFH, 0FFH,
                                   0FFH, 0FFH, 0FFH, 0FFH);
TYPE
   TPtr  = POINTER TO Nodo;
   TSerp = POINTER TO RECORD
              cabeza, cola : TPtr;
           END;
   Nodo  = RECORD
              pos : Graf.TPos;
              sig : TPtr;
           END;

PROCEDURE Crear(VAR s : TSerp; pos : Graf.TPos);
BEGIN
   NEW(s);
   NEW(s^.cabeza);
   s^.cabeza^.pos := pos;
   s^.cola := s^.cabeza;
   s^.cabeza^.sig := NIL;
END Crear;

PROCEDURE Eliminar(VAR ptr : TPtr);
VAR
   ret : TPtr;
BEGIN
   WHILE (ptr <> NIL) DO
      ret := ptr;
      ptr := ptr^.sig;
      DISPOSE(ret);
   END;
END Eliminar;

PROCEDURE Destruir(VAR s : TSerp);
VAR
   ptr : TPtr;
BEGIN
   IF (s <> NIL) AND (s^.cola <> NIL) THEN
      ptr := s^.cola;
      Eliminar(ptr);
      DISPOSE(s);
   END;
END Destruir;

PROCEDURE VisualizarCabeza(s : TSerp);
BEGIN
   IF (s <> NIL) THEN
      Graph.SetFillMask(MASC_SERP);
      Graf.PintarCuadrado(s^.cabeza^.pos.x, s^.cabeza^.pos.y, COLOR_SERP);
   END;
END VisualizarCabeza;

PROCEDURE EsSerp(s : TSerp; pos : Graf.TPos) : BOOLEAN;
VAR
   ptr : TPtr;
   es  : BOOLEAN;
BEGIN
   IF (s <> NIL) THEN
      ptr := s^.cola;
      WHILE (ptr <> NIL) AND (pos <> ptr^.pos) DO
         ptr := ptr^.sig;
      END;
      RETURN (ptr <> NIL);
   ELSE
      RETURN FALSE;
   END;
END EsSerp;

PROCEDURE EsCola(s : TSerp; pos : Graf.TPos) : BOOLEAN;
VAR
   ptr : TPtr;
   es  : BOOLEAN;
BEGIN
   IF (s <> NIL) THEN
      ptr := s^.cola;
      WHILE (ptr <> NIL) AND (ptr^.sig <> s^.cabeza) AND (pos <> ptr^.pos) DO
         ptr := ptr^.sig;
      END;
      RETURN (pos = ptr^.pos);
   ELSE
      RETURN FALSE;
   END;
END EsCola;

PROCEDURE EsCabeza(s : TSerp; pos : Graf.TPos) : BOOLEAN;
BEGIN
   RETURN (PosicionCabeza(s) = pos)
END EsCabeza;

PROCEDURE MoverCabeza(VAR s : TSerp; mov : TMov);
VAR
   nuevo : TPtr;
BEGIN
   IF (s <> NIL) AND (s^.cabeza <> NIL) THEN
      NEW(nuevo);
      nuevo^.pos := PosicionCabeza(s);
      CASE mov OF
         arr : DEC(nuevo^.pos.y); |
         aba : INC(nuevo^.pos.y); |
         izq : DEC(nuevo^.pos.x); |
         der : INC(nuevo^.pos.x);
      END;
      s^.cabeza^.sig := nuevo;
      nuevo^.sig := NIL;
      s^.cabeza := nuevo;
   END;
END MoverCabeza;

PROCEDURE Longitud(s : TSerp) : CARDINAL;
VAR
   ptr  : TPtr;
   cont : CARDINAL;
BEGIN
   cont := 0;
   IF (s <> NIL) THEN
      ptr := s^.cola;
      WHILE (ptr <> NIL) DO
         ptr := ptr^.sig;
         INC(cont);
      END;
   END;
   RETURN cont;
END Longitud;

PROCEDURE PosicionCabeza(s : TSerp) : Graf.TPos;
BEGIN
   RETURN (s^.cabeza^.pos);
END PosicionCabeza;

PROCEDURE CortarCola(VAR s : TSerp; pos : Graf.TPos);
VAR
   ptr : TPtr;
BEGIN
   IF (s <> NIL) AND (s^.cola <> NIL) THEN
      WHILE (s^.cola^.pos <> pos) DO
         RecojerCola(s);
      END;
      ptr := s^.cola;
      s^.cola := s^.cola^.sig;
      DISPOSE(ptr);
   END;
END CortarCola;

PROCEDURE RecojerCola(VAR s : TSerp);
VAR
   ptr : TPtr;
BEGIN
   IF (s <> NIL) AND (s^.cola <> NIL) THEN
      Graph.SetFillMask(NORMAL);
      Graf.PintarCuadrado(s^.cola^.pos.x, s^.cola^.pos.y, NEGRO);
      ptr := s^.cola;
      s^.cola := s^.cola^.sig;
      DISPOSE(ptr);
   END;
END RecojerCola;

BEGIN
END Serp.
