(*  *)
(*   Autor:  Ignacio Prez Martn.  Fecha: 15 de Mayo de 2002.   *)
(*   Msica: Roco Lpez Abelaira.                               *)
(*  *)

MODULE Serpente;
IMPORT Graf, Fant, Serp, Pant, Comi, Music;
IMPORT Graph, IO, BiosIO, Lib, Window, FIO;
FROM Storage IMPORT ALLOCATE, DEALLOCATE;

CONST
   Arr = CHR(72);  LARGO        = 02;
   Aba = CHR(80);  MAX_COMI     = 12;
   Izq = CHR(75);  COLOR_CHOQUE = 10;
   Der = CHR(77);  MAX_CADENA   = 10;
   Esc = CHR(27);  MAX_RECORD   = 03;

   MASC_CHOQUE = Graph.FillMaskType(081H, 042H, 024H, 018H,
                                    018H, 024H, 042H, 081H);

TYPE
   TObjeto = (serp, fant, comi);
   LFant   = POINTER TO Nodo;
   Nodo    = RECORD
                f   : Fant.TFant;
                sig : LFant;
             END;
   TCadena = ARRAY [1..MAX_CADENA] OF CHAR;
   TCadNum = ARRAY [1..MAX_CADENA] OF CARDINAL;
   TRecord = ARRAY [1..MAX_RECORD] OF RECORD
                nombre : TCadNum;
                puntos : CARDINAL;
             END;
VAR
   DAy, D, DMus, D2 : CARDINAL;

PROCEDURE LimpiarBuffer;
VAR
   car : CHAR;
BEGIN
   WHILE IO.KeyPressed() DO
      car := IO.RdKey();
   END;
END LimpiarBuffer;

PROCEDURE CalcularPos(objeto : TObjeto) : Graf.TPos;
VAR
   pos : Graf.TPos;
BEGIN
   Lib.RANDOMIZE;
   CASE objeto OF
      serp : pos.x := Lib.RANDOM(10)+02; Lib.RANDOMIZE;
             pos.y := Lib.RANDOM(10)+03;                |
      fant : pos.x := Lib.RANDOM(15)+20; Lib.RANDOMIZE;
             pos.y := Lib.RANDOM(25)+03;                |
      comi : pos.x := Lib.RANDOM(40)+01; Lib.RANDOMIZE;
             pos.y := Lib.RANDOM(25)+03;
   END;
   RETURN pos;
END CalcularPos;

PROCEDURE EsFantasma(lf, act : LFant; pos : Graf.TPos) : BOOLEAN;
VAR
   ptr : LFant;
   es  : BOOLEAN;
BEGIN
   ptr := lf;
   es := FALSE;
   WHILE (ptr <> act) AND NOT es DO
      es := (Fant.Posicion(ptr^.f) = pos);
      ptr := ptr^.sig;
   END;
   RETURN es;
END EsFantasma;

PROCEDURE CrearFantasmas(VAR lf : LFant; num : CARDINAL; c : Comi.TComi);
VAR
   ptr : LFant;
   pos : Graf.TPos;
BEGIN
   IF (num > 0) THEN
      NEW(lf);
      REPEAT
         pos := CalcularPos(fant);
      UNTIL (Comi.Posicion(c) <> pos);
      Fant.Crear(lf^.f, pos);
      ptr := lf;
      DEC(num);
      WHILE (num > 0) DO
         NEW(ptr^.sig);
         ptr := ptr^.sig;
         REPEAT
            pos := CalcularPos(fant);
         UNTIL (Comi.Posicion(c) <> pos) AND NOT (EsFantasma(lf, ptr, pos));
         Fant.Crear(ptr^.f, pos);
         DEC(num);
      END;
      ptr := NIL;
   END;
END CrearFantasmas;

PROCEDURE DestruirFantasmas(VAR lf : LFant);
VAR
   ptr : LFant;
BEGIN
   WHILE (lf <> NIL) DO
      ptr := lf;
      lf := lf^.sig;
      Fant.Destruir(ptr^.f);
      DISPOSE(ptr);
   END;
END DestruirFantasmas;

PROCEDURE VisualizarFantasmas(lf : LFant);
VAR
   ptr : LFant;
BEGIN
   ptr := lf;
   WHILE (ptr <> NIL) DO
      Fant.Visualizar(ptr^.f);
      ptr := ptr^.sig;
   END;
END VisualizarFantasmas;

PROCEDURE MoverFantasmas(VAR lf : LFant; c : Comi.TComi);
VAR
   ptr : LFant;
   pos : Graf.TPos;
   num : CARDINAL;
   sigue : BOOLEAN;
BEGIN
   ptr := lf;
   Lib.RANDOMIZE;
   WHILE (ptr <> NIL) DO
      REPEAT
         pos := Fant.Posicion(ptr^.f);
         num := Lib.RANDOM(4);
         CASE num OF
            0 : DEC(pos.x); |
            1 : INC(pos.x); |
            2 : DEC(pos.y); |
            3 : INC(pos.y);
         END;
         sigue := NOT Pant.EsMuro(pos) AND (Comi.Posicion(c) <> pos);
         IF NOT sigue THEN
            num := (num+1) MOD 4;
            pos := Fant.Posicion(ptr^.f);
            CASE num OF
               0 : DEC(pos.x); |
               1 : INC(pos.x); |
               2 : DEC(pos.y); |
               3 : INC(pos.y);
            END;
         END;
         sigue := NOT Pant.EsMuro(pos) AND (Comi.Posicion(c) <> pos);
      UNTIL sigue AND
      NOT EsFantasma(lf, ptr, pos);
      Fant.Mover(ptr^.f, pos);
      ptr := ptr^.sig;
   END;
END MoverFantasmas;

PROCEDURE InicializarPartida(VAR s : Serp.TSerp; VAR c : Comi.TComi;
                             VAR lf : LFant; VAR cont, numcomi : CARDINAL;
                             VAR car : CHAR; VAR muerto : BOOLEAN;
                             numpantalla : CARDINAL);
VAR
   pos : Graf.TPos;
BEGIN
   Serp.Crear(s, CalcularPos(serp));
   REPEAT
      pos := CalcularPos(comi);
   UNTIL NOT Serp.EsSerp(s, pos);
   Comi.Crear(c, pos);
   CrearFantasmas(lf, numpantalla, c);
   Pant.Crear;
   Serp.VisualizarCabeza(s);
   Comi.Visualizar(c);
   VisualizarFantasmas(lf);
   cont    := LARGO;
   muerto  := FALSE;
   numcomi := 0;
   Music.MusicIni(DMus);
   car     := Der;
   LimpiarBuffer;
END InicializarPartida;

PROCEDURE FinalizarPartida(VAR s : Serp.TSerp; VAR c : Comi.TComi;
                             VAR lf : LFant);
BEGIN
   Serp.Destruir(s);
   Comi.Destruir(c);
   Pant.Destruir;
   DestruirFantasmas(lf);
END FinalizarPartida;

PROCEDURE Acabar(VAR muerto : BOOLEAN; pos : Graf.TPos; s : Serp.TSerp);
BEGIN
   Serp.RecojerCola(s);
   Graph.SetFillMask(MASC_CHOQUE);
   Graf.PintarCuadrado(pos.x, pos.y, COLOR_CHOQUE);
   muerto := TRUE;
END Acabar;

PROCEDURE ComeFantasma(lf : LFant; pos : Graf.TPos) : BOOLEAN;
VAR
   ptr  : LFant;
   come : BOOLEAN;
BEGIN
   ptr := lf;
   come := FALSE;
   WHILE (ptr <> NIL) AND NOT come DO
      come := (Fant.Posicion(ptr^.f) = pos);
      ptr := ptr^.sig;
   END;
   RETURN come;
END ComeFantasma;

PROCEDURE Come(cabeza : Graf.TPos; c : Comi.TComi) : BOOLEAN;
BEGIN
   RETURN (cabeza = Comi.Posicion(c));
END Come;

PROCEDURE ChocaFantasma(s : Serp.TSerp; lf : LFant;
                        VAR pos : Graf.TPos) : BOOLEAN;
VAR
   ptr   : LFant;
   choca : BOOLEAN;
BEGIN
   ptr := lf;
   choca := FALSE;
   WHILE (ptr <> NIL) AND NOT choca DO
      pos := Fant.Posicion(ptr^.f);
      choca := Serp.EsSerp(s, pos);
      ptr := ptr^.sig;
   END;
   RETURN choca;
END ChocaFantasma;

PROCEDURE HacerNueva(VAR c : Comi.TComi; s : Serp.TSerp; lf : LFant);
VAR
   pos : Graf.TPos;
BEGIN
   REPEAT
      pos := CalcularPos(comi);
   UNTIL NOT Pant.EsMuro(pos) AND NOT Serp.EsSerp(s, pos) AND
         NOT (EsFantasma(lf, NIL, pos));
   Comi.Hacer(c, pos);
   Comi.Visualizar(c);
END HacerNueva;

PROCEDURE Menu() : CHAR;
VAR
   opcion : CHAR;
BEGIN
   Graph.TextMode;
   Window.Clear;
   IO.WrLn; IO.WrLn; IO.WrLn; IO.WrLn;
   IO.WrStr("                              ͻ       "); IO.WrLn;
   IO.WrStr("                                SERPENTE        "); IO.WrLn;
   IO.WrStr("                              ͼ      "); IO.WrLn;
   IO.WrStr("                                       "); IO.WrLn;
   IO.WrLn;
   IO.WrStr("                            (J) Jugar Serpente.      "); IO.WrLn;
   IO.WrStr("                            (C) Cambiar Velocidad.   "); IO.WrLn;
   IO.WrStr("                            (R) Ver podio de Records."); IO.WrLn;
   IO.WrStr("                            (A) Acerca de Serpente.  "); IO.WrLn;
   IO.WrLn;
   IO.WrStr("                               [Esc] = Salir.");
   Window.CursorOff;
   REPEAT
      IO.RdLn;
      opcion := CAP(BiosIO.RdKey());
   UNTIL ((opcion = Esc) OR (opcion = "J") OR (opcion = "A")
           OR (opcion = "C") OR (opcion = "R"));
   Window.CursorOn;
   RETURN opcion;
END Menu;

PROCEDURE Acerca();
VAR
   car : CHAR;
BEGIN
   Window.Clear;
   IO.WrLn; IO.WrLn; IO.WrLn;
   Graf.WrDelay("       AUTOR:  Ignacio Prez Martn.   FECHA: Mlaga a 15 de Mayo de 2002.", D2); IO.WrLn;
   Graf.WrDelay("       MUSICA: Roco Lpez Abelaira.                     VERSION: 3.20 ()", D2);IO.WrLn; IO.WrLn;
   IO.WrLn;
   Graf.WrDelay("       DESCRIPCION:                                                       ", D2); IO.WrLn;
   IO.WrLn;
   Graf.WrDelay("                   Manejando a la serpiente con los cursores debes conse- ", D2); IO.WrLn;
   Graf.WrDelay("               guir la mayor puntuacin posible. La puntuacin depende del", D2); IO.WrLn;
   Graf.WrDelay("               nmero de pantalla, de la longitud de la serpiente y de los", D2);IO.WrLn;
   Graf.WrDelay("               cuadrados obtenidos.                                       ", D2);IO.WrLn;
   Graf.WrDelay("                   PAUSA = Cualquier tecla, despus cursores para continuar", D2);IO.WrLn;
   Graf.WrDelay("                           o ESC para acabar, pero perders una unidad de ", D2);IO.WrLn;
   Graf.WrDelay("                           cola (tomadlo como una represalia por cobarde).", D2);IO.WrLn;
   IO.WrLn;
   Graf.WrDelay("       DIFICULTADES:                                                      ", D2); IO.WrLn;
   IO.WrLn;
   Graf.WrDelay("                   No puedes chocar con los bordes, con la propia serpiente ", D2); IO.WrLn;
   Graf.WrDelay("               o los fantasmas. Si stos te dan en la cola, la perders.  ", D2); IO.WrLn; IO.WrLn; IO.WrLn;
   Graf.WrDelay("       Pulsa cualquier tecla para continuar con Serpente . . . ", D2);
   LimpiarBuffer;
   Window.CursorOff;
   car := IO.RdKey();
   Window.CursorOn;
END Acerca;

PROCEDURE CalcularPuntos(s : Serp.TSerp;
                         pantalla, comida : CARDINAL) : CARDINAL;
BEGIN
   RETURN ( (Serp.Longitud(s)*5)+(pantalla+1) + comida);
END CalcularPuntos;

PROCEDURE EscribirPuntos(puntos : CARDINAL);
VAR
   car : CHAR;
   a   : CARDINAL;
BEGIN
   Window.Clear;
   IO.WrLn; IO.WrLn; IO.WrLn; IO.WrLn; IO.WrLn; IO.WrLn;
   Graf.WrDelay("                              PUNTUACION: ", D2);
   IO.WrCard(puntos, 0);
   Graf.WrDelay(". ", D2);
END EscribirPuntos;

PROCEDURE LeerCadena(VAR nombre : TCadena);
VAR
   i : CARDINAL;
   c : CHAR;
BEGIN
   i := 1;
   IO.RdLn;
   c := IO.RdChar();
   WHILE (i <= MAX_CADENA) AND (c <> CHR(13)) DO
      nombre[i] := c;
      c := IO.RdChar();
      INC(i);
   END;
   IF (i <= MAX_CADENA) THEN
      nombre[i] := "@";
   END;
END LeerCadena;

PROCEDURE ConvertirNombre(nombre : TCadena) : TCadNum;
VAR
   cadnum : TCadNum;
   i      : CARDINAL;
BEGIN
   i := 1;
   WHILE (i <= MAX_CADENA) AND (nombre[i] <> "@") DO
      cadnum[i] := ORD(nombre[i])+100;
      INC(i);
   END;
   IF (i <= MAX_CADENA) THEN
      cadnum[i] := 247;
   END;
   RETURN cadnum;
END ConvertirNombre;

PROCEDURE EscribirNombre(nombre : TCadNum);
VAR
   i : CARDINAL;
BEGIN
   i := 1;
   WHILE (i <= MAX_CADENA) AND (nombre[i] <> 247) DO
      IO.WrChar(CHR(nombre[i]-100));
      INC(i);
   END;
END EscribirNombre;

PROCEDURE Longitud(nombre : TCadNum) : CARDINAL;
VAR
   i, cont : CARDINAL;
BEGIN
   cont := 0;
   i := 1;
   WHILE (i <= MAX_CADENA) AND (nombre[i] <> 247) DO
      INC(cont);
      INC(i);
   END;
   RETURN cont;
END Longitud;

PROCEDURE EscribirCabecera;
BEGIN
   Window.Clear;
   IO.WrLn; IO.WrLn; IO.WrLn; IO.WrLn; IO.WrLn; IO.WrLn;
   Graf.WrDelay("                                    RECORDS", D2); IO.WrLn; IO.WrLn;
   IO.WrStr("                       NOMBRE                   PUNTOS"); IO.WrLn;
   IO.WrLn;
END EscribirCabecera;

PROCEDURE VerPodioRecords;
VAR
   arch    : FIO.File;
   tabla   : TRecord;
   i, a, j : CARDINAL;
   car     : CHAR;
BEGIN
   IF FIO.Exists("RECORDS.IPM") THEN
      arch := FIO.Open("RECORDS.IPM");
      a := FIO.RdBin(arch, tabla, SIZE(TRecord));
      EscribirCabecera;
      FOR i:=1 TO MAX_RECORD DO
         IO.WrStr("                       ");
         EscribirNombre(tabla[i].nombre);
         IO.WrStr("               ");
         FOR j:=0 TO (10-Longitud(tabla[i].nombre)) DO
            IO.WrStr(" ");
         END;
         IO.WrCard(tabla[i].puntos, 5);
         IO.WrLn;
      END;
      FIO.Close(arch);
   ELSE
      IO.WrLn; IO.WrLn;
      Graf.WrDelay("                         No existen records en el podio", D2);
   END;
   LimpiarBuffer;
   Window.CursorOff;
   car := IO.RdKey();
   Window.CursorOn;
END VerPodioRecords;

PROCEDURE MeterRecord(puntos : CARDINAL);
VAR
   tabla   : TRecord;
   arch, arch2 : FIO.File;
   i, a, j : CARDINAL;
   car     : CHAR;
   nombre  : TCadena;
BEGIN
   IF FIO.Exists("RECORDS.IPM") THEN
      arch := FIO.Open("RECORDS.IPM");
      arch2 := FIO.Open("RECORDS.IPM");
      a := FIO.RdBin(arch2, tabla, SIZE(TRecord));
      i := 1;
      WHILE (i <= MAX_RECORD) AND (tabla[i].puntos > puntos) DO
         INC(i);
      END;
      IF (i <= MAX_RECORD) AND (tabla[i].puntos <> puntos) THEN
         IO.WrLn; IO.WrLn;
         Graf.WrDelay("                     Ha conseguido un nuevo Record!!", D2); IO.WrLn;
         j := MAX_RECORD-1;
         WHILE (j>=i) DO
            tabla[j+1].nombre := tabla[j].nombre;
            tabla[j+1].puntos := tabla[j].puntos;
            DEC(j);
         END;
         LimpiarBuffer;
         Graf.WrDelay("                       Introduzca su nombre: ", D2);
         LeerCadena(nombre);
         tabla[i].nombre := ConvertirNombre(nombre);
         tabla[i].puntos := puntos;
         IO.WrLn; Graf.WrDelay("                           Record almacenado!! ", D2);
         FIO.WrBin(arch, tabla, SIZE(TRecord));
         FIO.Close(arch);
         FIO.Close(arch2);
         LimpiarBuffer;
         Window.CursorOff;
         car := IO.RdKey();
         VerPodioRecords;
      ELSE
         Window.CursorOff;
         LimpiarBuffer;
         car := IO.RdKey();
      END;
   ELSE
      arch := FIO.Create("RECORDS.IPM");
      IO.WrLn; IO.WrLn;
      Graf.WrDelay   ("                     Ha conseguido un nuevo Record!!", D2); IO.WrLn;
      LimpiarBuffer;
      Graf.WrDelay   ("                       Introduzca su nombre: ", D2);
      LeerCadena(nombre);
      tabla[1].nombre := ConvertirNombre(nombre);
      tabla[1].puntos := puntos;
      IO.WrLn; Graf.WrDelay("                           Record almacenado!! ", D2);
      FOR i:=2 TO MAX_RECORD DO
         tabla[i].nombre := ConvertirNombre("----------");
         tabla[i].puntos := 0;
      END;
      FIO.WrBin(arch, tabla, SIZE(TRecord));
      FIO.Close(arch);
      LimpiarBuffer;
      Window.CursorOff;
      car := IO.RdKey();
      VerPodioRecords;
   END;
   Window.CursorOn;
END MeterRecord;

PROCEDURE CambiarVelocidad;
VAR
  num : CARDINAL;
BEGIN
   Window.Clear;
   IO.WrLn; IO.WrLn; IO.WrLn; IO.WrLn; IO.WrLn; IO.WrLn;
   IO.WrStr("       La velocidad depende del ordenador. A mayor num menor velocidad. "); IO.WrLn; IO.WrLn;
   Graf.WrDelay("                             VELOCIDAD (de 1 a 25): ", D2);
   LimpiarBuffer;
   num := IO.RdCard();
   D  := num*40;     DAy  := D;
   D2 := D DIV 2;    DMus := (D*10) DIV 3;
END CambiarVelocidad;

VAR
   s    : Serp.TSerp;  car, ant, opcion : CHAR; poschoque       : Graf.TPos;
   c    : Comi.TComi;  muerto   : BOOLEAN;      numcomi, puntos : CARDINAL;
   lf   : LFant;       cabeza   : Graf.TPos;    numpantalla     : CARDINAL;
   cont : CARDINAL;
BEGIN
   Window.TextColor(Window.LightCyan);
   CambiarVelocidad;
   IO.WrLn;
   Graf.WrDelay(" SERPENTE v3.20 ()", D2);
   opcion := Menu();
   WHILE (opcion <> Esc) DO
   IF (opcion = "J") THEN
   numpantalla := 0;
   puntos := 0;
   InicializarPartida(s, c, lf, cont, numcomi, car, muerto, numpantalla);
   WHILE (car <> Esc) AND NOT muerto DO
      IO.RdLn;
      ant := car;
      IF BiosIO.KeyPressed() THEN
         car := BiosIO.RdKey();
         car := BiosIO.RdKey();
         IF (car <> Arr) AND (car <> Aba) AND (car <> Izq) AND (car <> Der)
         AND (car <> Esc) THEN
            car := ant;
            car := BiosIO.RdKey();
         END;
      END;
      CASE car OF
         Arr : Serp.MoverCabeza(s, Serp.arr); |
         Aba : Serp.MoverCabeza(s, Serp.aba); |
         Izq : Serp.MoverCabeza(s, Serp.izq); |
         Der : Serp.MoverCabeza(s, Serp.der);
      END;
      MoverFantasmas(lf, c);
      cabeza := Serp.PosicionCabeza(s);
      IF Pant.EsMuro(cabeza) OR ComeFantasma(lf, cabeza)
      OR Serp.EsCola(s, cabeza) THEN
         Acabar(muerto, cabeza, s);
      ELSIF Come(cabeza, c) THEN
         Music.Beep(DAy);
         Comi.Comer(c);
         Pant.IncPuntos(numcomi);
         INC(cont, LARGO);
         IF (numcomi = MAX_COMI) THEN
            puntos := puntos + CalcularPuntos(s, numpantalla, numcomi);
            FinalizarPartida(s, c, lf);
            INC(numpantalla, 2);
            InicializarPartida(s, c, lf, cont, numcomi, car, muerto,
                               numpantalla);
         ELSE
            HacerNueva(c, s, lf);
         END;
      ELSIF ChocaFantasma(s, lf, poschoque) THEN
         Music.Ay(DAy);
         Serp.CortarCola(s, poschoque);
      ELSIF (cont = 0) THEN
         Serp.RecojerCola(s);
      ELSE
         DEC(cont);
      END;
      IF NOT muerto THEN
         Serp.VisualizarCabeza(s);
         VisualizarFantasmas(lf);
      END;
      Lib.Delay(D);
   END;
   Music.MusicFin(DMus);
   puntos := puntos + CalcularPuntos(s, numpantalla, numcomi);
   FinalizarPartida(s, c, lf);
   EscribirPuntos(puntos);
   MeterRecord(puntos);
   ELSIF (opcion = "C") THEN
      CambiarVelocidad;
   ELSIF (opcion = "R") THEN
      VerPodioRecords;
   ELSE
      Acerca;
   END;
   opcion := Menu();
   END;
   Window.TextColor(Window.White);
   Window.Clear;
   IO.WrLn;
   Graf.WrDelay(" Sugerencias a djnash_@hotmail.com", D2);
   LimpiarBuffer;
END Serpente.
